class GenericNodeMixin(object):

    def span_event(
            self, base_attrs=None, parent_guid=None):
        i_attrs = base_attrs and base_attrs.copy() or {}
        i_attrs['type'] = 'Span'
        i_attrs['name'] = self.name
        i_attrs['guid'] = self.guid
        i_attrs['timestamp'] = int(self.start_time * 1000)
        i_attrs['duration'] = self.duration
        i_attrs['category'] = 'generic'

        if parent_guid:
            i_attrs['parentId'] = parent_guid

        return [i_attrs, {}, {}]  # intrinsics, agent attrs, user attrs

    def span_events(self,
            stats, base_attrs=None, parent_guid=None):

        yield self.span_event(
                base_attrs=base_attrs,
                parent_guid=parent_guid)

        for child in self.children:
            for event in child.span_events(
                    stats,
                    base_attrs=base_attrs,
                    parent_guid=self.guid):
                yield event


class DatastoreNodeMixin(GenericNodeMixin):

    @property
    def name(self):
        product = self.product
        target = self.target
        operation = self.operation or 'other'

        if target:
            name = 'Datastore/statement/%s/%s/%s' % (product, target,
                    operation)
        else:
            name = 'Datastore/operation/%s/%s' % (product, operation)

        return name

    def span_event(self, *args, **kwargs):
        attrs = super(DatastoreNodeMixin, self).span_event(*args, **kwargs)
        i_attrs = attrs[0]

        i_attrs['category'] = 'datastore'
        i_attrs['component'] = self.product
        i_attrs['db.instance'] = self.database_name or 'Unknown'
        i_attrs['peer.address'] = '%s:%s' % (
                self.instance_hostname or 'Unknown',
                self.port_path_or_id or 'Unknown')
        i_attrs['peer.hostname'] = self.instance_hostname or 'Unknown'
        i_attrs['span.kind'] = 'client'

        return attrs


class ExternalNodeMixin(GenericNodeMixin):

    @property
    def name(self):
        return 'External/%s/%s/%s' % (
                self.netloc, self.library, self.method or '')

    def span_event(self, *args, **kwargs):
        attrs = super(ExternalNodeMixin, self).span_event(*args, **kwargs)
        i_attrs = attrs[0]

        i_attrs['category'] = 'http'
        i_attrs['span.kind'] = 'client'
        i_attrs['http.url'] = self.url_with_path
        i_attrs['component'] = self.library

        if self.method:
            i_attrs['http.method'] = self.method

        return attrs



from newrelic.common.object_wrapper import wrap_object
from newrelic.api.database_trace import register_database_client

from newrelic.hooks.database_dbapi2 import ConnectionFactory


def instrument_ibm_db_dbi(module):
    register_database_client(module, database_product='IBMDB2',
            quoting_style='single', explain_query='EXPLAIN',
            explain_stmts=('select', 'insert', 'update', 'delete'))

    wrap_object(module, 'connect', ConnectionFactory, (module,))



from collections import namedtuple

import newrelic.core.trace_node

from newrelic.core.node_mixin import GenericNodeMixin
from newrelic.core.metric import TimeMetric

_MemcacheNode = namedtuple('_MemcacheNode',
        ['command', 'children', 'start_time', 'end_time', 'duration',
        'exclusive', 'is_async', 'guid'])


class MemcacheNode(_MemcacheNode, GenericNodeMixin):

    @property
    def name(self):
        return 'Memcache/%s' % self.command

    def time_metrics(self, stats, root, parent):
        """Return a generator yielding the timed metrics for this
        memcache node as well as all the child nodes.

        """

        yield TimeMetric(name='Memcache/all', scope='',
                duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebTransaction':
            yield TimeMetric(name='Memcache/allWeb', scope='',
                    duration=self.duration, exclusive=self.exclusive)
        else:
            yield TimeMetric(name='Memcache/allOther', scope='',
                    duration=self.duration, exclusive=self.exclusive)

        name = 'Memcache/%s' % self.command

        yield TimeMetric(name=name, scope='', duration=self.duration,
                  exclusive=self.exclusive)

        yield TimeMetric(name=name, scope=root.path,
                duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, stats, root, connections):
        name = root.string_table.cache(self.name)

        start_time = newrelic.core.trace_node.node_start_time(root, self)
        end_time = newrelic.core.trace_node.node_end_time(root, self)

        children = []

        root.trace_node_count += 1

        params = {}
        params['exclusive_duration_millis'] = 1000.0 * self.exclusive

        return newrelic.core.trace_node.TraceNode(start_time=start_time,
                end_time=end_time, name=name, params=params, children=children,
                label=None)



import logging
import os
import string
import re

from newrelic.packages import requests
from newrelic.core.internal_metrics import internal_count_metric


_logger = logging.getLogger(__name__)
VALID_CHARS_RE = re.compile(r'[0-9a-zA-Z_ ./-]')


class CommonUtilization(object):
    METADATA_URL = ''
    HEADERS = None
    EXPECTED_KEYS = ()
    VENDOR_NAME = ''
    TIMEOUT = 0.5

    @classmethod
    def record_error(cls, resource, data):
        # As per spec
        internal_count_metric(
                'Supportability/utilization/%s/error' % cls.VENDOR_NAME, 1)
        _logger.warning('Invalid %r data (%r): %r',
                cls.VENDOR_NAME, resource, data)

    @classmethod
    def fetch(cls):
        # Create own requests session and disable all environment variables,
        # so that we can bypass any proxy set via env var for this request.

        session = requests.Session()
        session.trust_env = False

        try:
            resp = session.get(cls.METADATA_URL, timeout=cls.TIMEOUT,
                    headers=cls.HEADERS)
            resp.raise_for_status()
        except Exception as e:
            resp = None
            _logger.debug('Unable to fetch %s data from %r: %r',
                    cls.VENDOR_NAME, cls.METADATA_URL, e)

        return resp

    @classmethod
    def get_values(cls, response):
        if response is None:
            return

        try:
            j = response.json()
        except ValueError:
            _logger.debug('Invalid %s data (%r): %r',
                    cls.VENDOR_NAME, cls.METADATA_URL, response.text)
            return

        return j

    @classmethod
    def valid_chars(cls, data):
        if data is None:
            return False

        for c in data:
            if not VALID_CHARS_RE.match(c) and ord(c) < 0x80:
                return False

        return True

    @classmethod
    def valid_length(cls, data):
        if data is None:
            return False

        b = data.encode('utf-8')
        valid = len(b) <= 255
        if valid:
            return True

        return False

    @classmethod
    def normalize(cls, key, data):
        if data is None:
            return

        try:
            stripped = data.strip()

            if (stripped and cls.valid_length(stripped) and
                    cls.valid_chars(stripped)):
                return stripped
        except:
            pass

    @classmethod
    def sanitize(cls, values):
        if values is None:
            return

        out = {}
        for key in cls.EXPECTED_KEYS:
            metadata = values.get(key, None)
            if not metadata:
                cls.record_error(key, metadata)
                return

            normalized = cls.normalize(key, metadata)
            if not normalized:
                cls.record_error(key, metadata)
                return

            out[key] = normalized

        return out

    @classmethod
    def detect(cls):
        response = cls.fetch()
        values = cls.get_values(response)
        return cls.sanitize(values)


class AWSUtilization(CommonUtilization):
    EXPECTED_KEYS = ('availabilityZone', 'instanceId', 'instanceType')
    METADATA_URL = '%s/2016-09-02/dynamic/instance-identity/document' % (
        'http://169.254.169.254'
    )
    VENDOR_NAME = 'aws'


class AzureUtilization(CommonUtilization):
    METADATA_URL = ('http://169.254.169.254'
            '/metadata/instance/compute?api-version=2017-03-01')
    EXPECTED_KEYS = ('location', 'name', 'vmId', 'vmSize')
    HEADERS = {'Metadata': 'true'}
    VENDOR_NAME = 'azure'


class GCPUtilization(CommonUtilization):
    EXPECTED_KEYS = ('id', 'machineType', 'name', 'zone')
    HEADERS = {'Metadata-Flavor': 'Google'}
    METADATA_URL = 'http://%s/computeMetadata/v1/instance/?recursive=true' % (
            'metadata.google.internal')
    VENDOR_NAME = 'gcp'

    @classmethod
    def normalize(cls, key, data):
        if data is None:
            return

        if key in ('machineType', 'zone'):
            formatted = data.strip().split('/')[-1]
        elif key == 'id':
            formatted = str(data)
        else:
            formatted = data

        return super(GCPUtilization, cls).normalize(key, formatted)


class PCFUtilization(CommonUtilization):
    EXPECTED_KEYS = ('cf_instance_guid', 'cf_instance_ip', 'memory_limit')
    VENDOR_NAME = 'pcf'

    @staticmethod
    def fetch():
        cf_instance_guid = os.environ.get('CF_INSTANCE_GUID')
        cf_instance_ip = os.environ.get('CF_INSTANCE_IP')
        memory_limit = os.environ.get('MEMORY_LIMIT')
        pcf_vars = (cf_instance_guid, cf_instance_ip, memory_limit)
        if all(pcf_vars):
            return pcf_vars

    @classmethod
    def get_values(cls, response):
        if response is None or len(response) != 3:
            return

        values = {}
        for k, v in zip(cls.EXPECTED_KEYS, response):
            if hasattr(v, 'decode'):
                v = v.decode('utf-8')
            values[k] = v
        return values


class DockerUtilization(CommonUtilization):
    VENDOR_NAME = 'docker'
    EXPECTED_KEYS = ('id',)
    METADATA_FILE = '/proc/self/cgroup'
    DOCKER_RE = re.compile(r'([0-9a-f]{64,})')

    @classmethod
    def fetch(cls):
        try:
            with open(cls.METADATA_FILE, 'rb') as f:
                for line in f:
                    stripped = line.decode('utf-8').strip()
                    cgroup = stripped.split(':')
                    if len(cgroup) != 3:
                        continue
                    subsystems = cgroup[1].split(',')
                    if 'cpu' in subsystems:
                        return cgroup[2]
        except:
            # There are all sorts of exceptions that can occur here
            # (i.e. permissions, non-existent file, etc)
            pass

    @classmethod
    def get_values(cls, contents):
        if contents is None:
            return

        value = contents.split('/')[-1]
        match = cls.DOCKER_RE.search(value)
        if match:
            value = match.group(0)
            return {'id': value}

    @classmethod
    def valid_chars(cls, data):
        if data is None:
            return False

        hex_digits = set(string.hexdigits)

        valid = all((c in hex_digits for c in data))
        if valid:
            return True

        return False

    @classmethod
    def valid_length(cls, data):
        if data is None:
            return False

        # Must be exactly 64 characters
        valid = len(data) == 64
        if valid:
            return True

        return False



from __future__ import print_function

from newrelic.admin import command, usage

@command('license-info', '',
"""Prints out the LICENSE for the New Relic Python Agent.""")
def license_info(args):
    import os
    import sys

    if len(args) != 0:
        usage('license-info')
        sys.exit(1)

    from newrelic import __file__ as package_root
    package_root = os.path.dirname(package_root)

    license_file = os.path.join(package_root, 'LICENSE')

    license = open(license_file, 'r').read()

    print(license, end='')



import asyncio
import sys

from newrelic.api.application import application_instance
from newrelic.api.web_transaction import WebTransaction
from newrelic.api.transaction import current_transaction
from newrelic.api.function_trace import function_trace
from newrelic.common.object_wrapper import (wrap_function_wrapper, ObjectProxy,
    function_wrapper)
from newrelic.common.object_names import callable_name
from newrelic.core.config import ignore_status_code


class NRTransactionCoroutineWrapper(ObjectProxy):
    def __init__(self, wrapped, request):
        super(NRTransactionCoroutineWrapper, self).__init__(wrapped)
        self._nr_transaction = None
        self._nr_request = request

        environ = {
            'PATH_INFO': request.path,
            'REQUEST_METHOD': request.method,
            'CONTENT_TYPE': request.content_type,
            'QUERY_STRING': request.query_string,
        }
        for k, v in request.headers.items():
            normalized_key = k.replace('-', '_').upper()
            http_key = 'HTTP_%s' % normalized_key
            environ[http_key] = v
        self._nr_environ = environ

    def __iter__(self):
        return self

    def __await__(self):
        return self

    def __next__(self):
        return self.send(None)

    def send(self, value):
        if not self._nr_transaction:
            # create and start the transaction
            app = application_instance()
            txn = WebTransaction(app, self._nr_environ)

            import sanic
            txn.add_framework_info(
                    name='Sanic', version=sanic.__version__)

            self._nr_transaction = txn

            if txn.enabled:
                txn.__enter__()
                txn.drop_transaction()

        txn = self._nr_transaction

        # transaction may not be active
        if not txn.enabled:
            return self.__wrapped__.send(value)

        txn.save_transaction()

        try:
            r = self.__wrapped__.send(value)
            txn.drop_transaction()
            return r
        except (GeneratorExit, StopIteration):
            self._nr_transaction.__exit__(None, None, None)
            self._nr_request = None
            raise
        except:
            self._nr_transaction.__exit__(*sys.exc_info())
            self._nr_request = None
            raise

    def throw(self, *args, **kwargs):
        txn = self._nr_transaction

        # transaction may not be active
        if not txn.enabled:
            return self.__wrapped__.throw(*args, **kwargs)

        txn.save_transaction()
        try:
            r = self.__wrapped__.throw(*args, **kwargs)
            txn.drop_transaction()
            return r
        except (GeneratorExit, StopIteration):
            self._nr_transaction.__exit__(None, None, None)
            self._nr_request = None
            raise
        except asyncio.CancelledError:
            self._nr_transaction.ignore_transaction = True
            self._nr_transaction.__exit__(None, None, None)
            self._nr_request = None
            raise
        except:
            self._nr_transaction.__exit__(*sys.exc_info())
            self._nr_request = None
            raise

    def close(self):
        txn = self._nr_transaction

        # transaction may not be active
        if not txn.enabled:
            return self.__wrapped__.close()

        txn.save_transaction()
        try:
            r = self.__wrapped__.close()
            self._nr_transaction.__exit__(None, None, None)
            self._nr_request = None
            return r
        except:
            self._nr_transaction.__exit__(*sys.exc_info())
            self._nr_request = None
            raise


def _bind_request(request, *args, **kwargs):
    return request


def _nr_sanic_transaction_wrapper_(wrapped, instance, args, kwargs):
    # get the coroutine
    coro = wrapped(*args, **kwargs)
    request = _bind_request(*args, **kwargs)

    # Wrap the coroutine
    return NRTransactionCoroutineWrapper(coro, request)


def _bind_add(uri, methods, handler, *args, **kwargs):
    return uri, methods, handler, args, kwargs


@function_wrapper
def _nr_wrapper_handler_(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)
    transaction.set_transaction_name(name, priority=3)

    return function_trace(name=name)(wrapped)(*args, **kwargs)


@function_wrapper
def _nr_sanic_router_add(wrapped, instance, args, kwargs):
    uri, methods, handler, args, kwargs = _bind_add(*args, **kwargs)

    # Cache the callable_name on the handler object
    callable_name(handler)
    wrapped_handler = _nr_wrapper_handler_(handler)

    return wrapped(uri, methods, wrapped_handler, *args, **kwargs)


@function_wrapper
def _nr_sanic_router_get(wrapped, instance, args, kwargs):
    # Rename all transactions that generate an exception in the router get
    try:
        return wrapped(*args, **kwargs)
    except Exception:
        transaction = current_transaction()
        if transaction:
            name = callable_name(wrapped)
            transaction.set_transaction_name(name, priority=2)
        raise


@function_wrapper
def _nr_wrapper_error_handler_(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if not transaction:
        return wrapped(*args, **kwargs)

    name = callable_name(wrapped)
    transaction.set_transaction_name(name, priority=1)
    try:
        response = function_trace(name=name)(wrapped)(*args, **kwargs)
    except:
        transaction.record_exception()
        raise

    return response


def _bind_error_add(exception, handler, *args, **kwargs):
    return exception, handler


@function_wrapper
def _nr_sanic_error_handlers(wrapped, instance, args, kwargs):
    exception, handler = _bind_error_add(*args, **kwargs)

    # Cache the callable_name on the handler object
    callable_name(handler)
    wrapped_handler = _nr_wrapper_error_handler_(handler)

    return wrapped(exception, wrapped_handler)


@function_wrapper
def error_response(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    exc_info = sys.exc_info()
    try:
        response = wrapped(*args, **kwargs)
    except:
        transaction.record_exception(*exc_info)
        # We record the original exception and the exception generated by the
        # error handler
        transaction.record_exception()
        raise
    else:
        # response can be a response object or a coroutine
        if hasattr(response, 'status'):
            if not ignore_status_code(response.status):
                transaction.record_exception(*exc_info)
        else:
            transaction.record_exception(*exc_info)
    finally:
        exc_info = None

    return response


def _sanic_app_init(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)

    error_handler = getattr(instance, 'error_handler')
    if hasattr(error_handler, 'response'):
        instance.error_handler.response = error_response(
                error_handler.response)
    if hasattr(error_handler, 'add'):
        error_handler.add = _nr_sanic_error_handlers(
                error_handler.add)

    router = getattr(instance, 'router')
    if hasattr(router, 'add'):
        router.add = _nr_sanic_router_add(router.add)
    if hasattr(router, 'get'):
        # Cache the callable_name on the router.get
        callable_name(router.get)
        router.get = _nr_sanic_router_get(router.get)

    return result


def _nr_sanic_response_parse_headers(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    # instance is the response object
    cat_headers = transaction.process_response(str(instance.status),
            instance.headers.items())

    for header_name, header_value in cat_headers:
        if header_name not in instance.headers:
            instance.headers[header_name] = header_value

    return wrapped(*args, **kwargs)


def _nr_wrapper_middleware_(attach_to):
    is_request_middleware = attach_to == 'request'

    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        name = callable_name(wrapped)
        if is_request_middleware:
            transaction.set_transaction_name(name, priority=2)
        response = function_trace(name=name)(wrapped)(*args, **kwargs)

        return response

    return _wrapper


def _bind_middleware(middleware, attach_to='request', *args, **kwargs):
    return middleware, attach_to


def _nr_sanic_register_middleware_(wrapped, instance, args, kwargs):
    middleware, attach_to = _bind_middleware(*args, **kwargs)

    # Cache the callable_name on the middleware object
    callable_name(middleware)
    wrapped_middleware = _nr_wrapper_middleware_(attach_to)(middleware)
    wrapped(wrapped_middleware, attach_to)
    return middleware


def instrument_sanic_app(module):
    wrap_function_wrapper(module, 'Sanic.handle_request',
        _nr_sanic_transaction_wrapper_)
    wrap_function_wrapper(module, 'Sanic.__init__',
        _sanic_app_init)
    wrap_function_wrapper(module, 'Sanic.register_middleware',
        _nr_sanic_register_middleware_)


def instrument_sanic_response(module):
    wrap_function_wrapper(module, 'BaseHTTPResponse._parse_headers',
        _nr_sanic_response_parse_headers)



import functools

from newrelic.api.coroutine_trace import return_value_fn
from newrelic.api.time_trace import TimeTrace
from newrelic.api.transaction import current_transaction
from newrelic.core.memcache_node import MemcacheNode
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object


class MemcacheTrace(TimeTrace):

    node = MemcacheNode

    def __init__(self, transaction, command):
        super(MemcacheTrace, self).__init__(transaction)

        self.command = command

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, dict(
                command=self.command))

    def terminal_node(self):
        return True


def MemcacheTraceWrapper(wrapped, command):

    return_value = return_value_fn(wrapped)

    def _nr_wrapper_memcache_trace_(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        if callable(command):
            if instance is not None:
                _command = command(instance, *args, **kwargs)
            else:
                _command = command(*args, **kwargs)
        else:
            _command = command

        trace = MemcacheTrace(transaction, _command)
        return return_value(trace, lambda: wrapped(*args, **kwargs))

    return FunctionWrapper(wrapped, _nr_wrapper_memcache_trace_)


def memcache_trace(command):
    return functools.partial(MemcacheTraceWrapper, command=command)


def wrap_memcache_trace(module, object_path, command):
    wrap_object(module, object_path, MemcacheTraceWrapper, (command,))



from newrelic.api.datastore_trace import wrap_datastore_trace

_pysolr_client_methods = ('search', 'more_like_this', 'suggest_terms', 'add',
'delete', 'commit', 'optimize', 'extract')

_pysolr_admin_methods = ('status', 'create', 'reload', 'rename', 'swap',
    'unload', 'load')

def instrument_pysolr(module):
    for name in _pysolr_client_methods:
        if hasattr(module.Solr, name):
            wrap_datastore_trace(module.Solr, name,
                    product='Solr', target=None, operation=name)

    if hasattr(module, 'SolrCoreAdmin'):
        for name in _pysolr_admin_methods:
            if hasattr(module.SolrCoreAdmin, name):
                wrap_datastore_trace(module.SolrCoreAdmin, name,
                        product='Solr', target=None,
                        operation='admin.%s' % name)



from collections import namedtuple

import newrelic.core.trace_node

from newrelic.core.node_mixin import GenericNodeMixin
from newrelic.core.metric import TimeMetric

_SolrNode = namedtuple('_SolrNode',
        ['library', 'command', 'children', 'start_time', 'end_time',
        'duration', 'exclusive', 'guid', 'is_async'])


class SolrNode(_SolrNode, GenericNodeMixin):

    @property
    def name(self):
        return 'SolrClient/%s/%s' % (self.library, self.command)

    def time_metrics(self, stats, root, parent):
        """Return a generator yielding the timed metrics for this
        memcache node as well as all the child nodes.

        """
        yield TimeMetric(name='Solr/all', scope='',
                duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebTransaction':
            yield TimeMetric(name='Solr/allWeb', scope='',
                    duration=self.duration, exclusive=self.exclusive)
        else:
            yield TimeMetric(name='Solr/allOther', scope='',
                    duration=self.duration, exclusive=self.exclusive)

        name = 'Solr/%s' % self.command

        yield TimeMetric(name=name, scope='', duration=self.duration,
                  exclusive=self.exclusive)

        yield TimeMetric(name=name, scope=root.path,
                duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, stats, root, connections):
        name = root.string_table.cache(self.name)

        start_time = newrelic.core.trace_node.node_start_time(root, self)
        end_time = newrelic.core.trace_node.node_end_time(root, self)

        children = []

        root.trace_node_count += 1

        params = {}
        params['exclusive_duration_millis'] = 1000.0 * self.exclusive

        return newrelic.core.trace_node.TraceNode(start_time=start_time,
                end_time=end_time, name=name, params=params, children=children,
                label=None)



from collections import namedtuple

_XraySession = namedtuple('XraySession',
        ['xray_id', 'key_txn', 'stop_time_s', 'max_traces',
            'sample_period_s'])

class XraySession(_XraySession):

    def get_trace_count(self):
        if getattr(self, '_trace_count', None) is None:
            self._trace_count = 0
        return self._trace_count

    def set_trace_count(self, count):
        self._trace_count = count

    trace_count = property(get_trace_count, set_trace_count)



import inspect
import logging

from newrelic.common.object_wrapper import ObjectProxy
from newrelic.common.object_names import callable_name

_logger = logging.getLogger(__name__)


if hasattr(inspect, 'iscoroutinefunction'):
    def is_coroutine_function(wrapped):
        return (inspect.iscoroutinefunction(wrapped) or
                inspect.isgeneratorfunction(wrapped))
else:
    def is_coroutine_function(wrapped):
        return inspect.isgeneratorfunction(wrapped)


def _iscoroutinefunction_tornado(fn):
    return hasattr(fn, '__tornado_coroutine__')


class TraceContext(object):
    # Assumption made for this context manager: no other object maintains a
    # reference to trace after the TraceContext is constructed.
    # This is important because the assumption made by the logic below is that
    # only the TraceContext object can call __enter__ on the trace (which may
    # delete the transaction).

    def __init__(self, trace):
        if not trace.transaction:
            self.trace = None
        else:
            self.trace = trace

        self.current_trace = None

    def __enter__(self):
        if not self.trace:
            return self

        self.current_trace = self.trace.transaction.current_node
        if not self.trace.activated:
            self.trace.__enter__()

            # Transaction can be cleared by calling enter (such as when tracing
            # as a child of a terminal node)
            # In this case, we clear out the transaction and defer to the
            # wrapped function
            if not self.trace.transaction:
                self.trace = None
        else:
            # In the case that the function trace is already active, notify the
            # transaction that this coroutine is now the current node (for
            # automatic parenting)
            self.trace.transaction.current_node = self.trace

        return self

    def __exit__(self, exc, value, tb):
        if not self.trace:
            return

        txn = self.trace.transaction

        # If the current node has been changed, record this as an error in a
        # supportability metric.
        if txn.current_node is not self.trace:
            txn.record_custom_metric(
                    'Supportability/Python/TraceContext/ExitNodeMismatch',
                    {'count': 1})
            _logger.debug('Trace context exited with an unexpected current '
                    'node. Current trace is %r. Expected trace is %r. '
                    'Please report this issue to New Relic Support.',
                    self.current_trace, self.trace)

        if exc in (StopIteration, GeneratorExit):
            self.trace.__exit__(None, None, None)
            self.trace = None
        elif exc:
            self.trace.__exit__(exc, value, tb)
            self.trace = None

        # Since the coroutine is returning control to the parent at this
        # point, we should notify the transaction that this coroutine is no
        # longer the current node for parenting purposes
        if self.current_trace:
            txn.current_node = self.current_trace

            # Clear out the current trace so that it cannot be reused in future
            # exits and doesn't maintain a dangling reference to a trace.
            self.current_trace = None

    def __del__(self):
        # If the trace goes out of scope, it's possible it's still active.
        # It's important that we end the trace so that the trace is reported.
        # It's also important that we don't change the current node as part of
        # this reporting.
        if self.trace and self.trace.activated:
            txn = self.trace.transaction
            if txn:
                current_trace = txn.current_node
                self.trace.__exit__(None, None, None)
                if current_trace is not self.trace:
                    txn.current_node = current_trace


class CoroutineTrace(ObjectProxy):
    def __init__(self, wrapped, trace):

        self._nr_trace_context = TraceContext(trace)

        # get the coroutine
        coro = wrapped()

        # Wrap the coroutine
        super(CoroutineTrace, self).__init__(coro)

    def __iter__(self):
        return self

    def __await__(self):
        return self

    def __next__(self):
        return self.send(None)

    next = __next__

    def send(self, value):
        with self._nr_trace_context:
            return self.__wrapped__.send(value)

    def throw(self, *args, **kwargs):
        with self._nr_trace_context:
            return self.__wrapped__.throw(*args, **kwargs)

    def close(self):
        try:
            with self._nr_trace_context:
                result = self.__wrapped__.close()
        except:
            raise
        else:
            # If the trace hasn't been exited, then exit the trace
            if self._nr_trace_context.trace:
                self._nr_trace_context.trace.__exit__(None, None, None)
                self._nr_trace_context.trace = None
            return result


def return_value_fn(wrapped):
    if _iscoroutinefunction_tornado(wrapped):
        if hasattr(wrapped, '__wrapped__'):
            coro_name = callable_name(wrapped.__wrapped__)
        else:
            coro_name = callable_name(wrapped)
        _logger.warning('The tornado coroutine function %r '
                '(tornado.gen.coroutine) has been incorrectly wrapped. To '
                'trace a tornado coroutine, the New Relic trace decorator '
                'must be the innermost decorator. For more information see '
                'https://docs.newrelic.com/docs/agents/python-agent/'
                'trace-decorators-tornado-coroutines', coro_name)

    if is_coroutine_function(wrapped):
        def return_value(trace, fn):
            return CoroutineTrace(fn, trace)
    else:
        def return_value(trace, fn):
            with trace:
                return fn()

    return return_value



import logging

from newrelic.hooks.framework_tornado_r3.util import (
        server_request_adapter_finish_finalize,
        server_request_adapter_on_connection_close_finalize)
from newrelic.common.object_wrapper import wrap_function_wrapper

_logger = logging.getLogger(__name__)

# For every request either finish() or on_connection_close() is called on the
# _ServerRequestAdapter. We require that one of these methods is called before
# the transaction is allowed to be finalized.


def _nr_wrapper__ServerRequestAdapter_on_connection_close_(wrapped, instance,
        args, kwargs):
    return server_request_adapter_on_connection_close_finalize(wrapped,
            instance, args, kwargs)


def _nr_wrapper__ServerRequestAdapter_finish_(wrapped, instance,
        args, kwargs):
    return server_request_adapter_finish_finalize(wrapped, instance, args,
            kwargs)


def _nr_wrapper__HTTPServer_start_request_(wrapped, instance,
        args, kwargs):

    delegate = wrapped(*args, **kwargs)

    wrap_function_wrapper(delegate, 'on_connection_close',
            _nr_wrapper__ServerRequestAdapter_on_connection_close_)
    wrap_function_wrapper(delegate, 'finish',
            _nr_wrapper__ServerRequestAdapter_finish_)

    return delegate


def instrument_tornado_httpserver(module):

    if hasattr(module, '_ServerRequestAdapter'):
        wrap_function_wrapper(module,
                '_ServerRequestAdapter.on_connection_close',
                _nr_wrapper__ServerRequestAdapter_on_connection_close_)
        wrap_function_wrapper(module, '_ServerRequestAdapter.finish',
                _nr_wrapper__ServerRequestAdapter_finish_)
    else:
        wrap_function_wrapper(module, 'HTTPServer.start_request',
                _nr_wrapper__HTTPServer_start_request_)



from newrelic.api.function_trace import FunctionTrace
from newrelic.common.object_wrapper import wrap_function_wrapper

from newrelic.hooks.framework_tornado import retrieve_current_transaction


def template_generate_wrapper(wrapped, instance, args, kwargs):
    transaction = retrieve_current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    with FunctionTrace(transaction, name=instance.name,
            group='Template/Render'):
        return wrapped(*args, **kwargs)


def template_generate_python_wrapper(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)

    if result is not None:
        return ('import newrelic.api.function_trace as _nr_fxn_trace\n'
                'import newrelic.api.transaction as _nr_txn\n'
                ) + result


def block_generate_wrapper(wrapped, instance, args, kwargs):

    def execute(writer, *args, **kwargs):
        if not hasattr(instance, 'line'):
            return wrapped(writer, *args, **kwargs)

        writer.write_line('with _nr_fxn_trace.FunctionTrace('
                '_nr_txn.current_transaction(), name=%r, '
                'group="Template/Block"):' % instance.name, instance.line)

        with writer.indent():
            writer.write_line("pass", instance.line)
            return wrapped(writer, *args, **kwargs)

    return execute(*args, **kwargs)


def instrument_tornado_template(module):
    wrap_function_wrapper(module, 'Template.generate',
            template_generate_wrapper)
    wrap_function_wrapper(module, 'Template._generate_python',
            template_generate_python_wrapper)
    wrap_function_wrapper(module, '_NamedBlock.generate',
            block_generate_wrapper)



from newrelic.packages import six

from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper

# An index name can be a string, None or a sequence. In the case of None
# an empty string or '*', it is the same as using '_all'. When a string
# it can also be a comma separated list of index names. A sequence
# obviously can also be more than one index name. Where we are certain
# there is only a single index name we use it, otherwise we use 'other'.

def _index_name(index):
    if not index or index == '*':
        return '_all'
    if not isinstance(index, six.string_types) or ',' in index:
        return 'other'
    return index

def _extract_kwargs_index(*args, **kwargs):
    return _index_name(kwargs.get('index'))

def _extract_args_index(index=None, *args, **kwargs):
    return _index_name(index)

def _extract_args_body_index(body=None, index=None, *args, **kwargs):
    return _index_name(index)

def _extract_args_doctype_body_index(doc_type=None, body=None, index=None,
        *args, **kwargs):
    return _index_name(index)

def _extract_args_field_index(field=None, index=None, *args, **kwargs):
    return _index_name(index)

def _extract_args_name_body_index(name=None, body=None, index=None,
        *args, **kwargs):
    return _index_name(index)

def _extract_args_name_index(name=None, index=None, *args, **kwargs):
    return _index_name(index)

def _extract_args_metric_index(metric=None, index=None, *args, **kwargs):
    return _index_name(index)

def wrap_elasticsearch_client_method(owner, name, arg_extractor, prefix=None):
    def _nr_wrapper_Elasticsearch_method_(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        # When arg_extractor is None, it means there is no target field
        # associated with this method. Hence this method will only
        # create an operation metric and no statement metric. This is
        # handled by setting the target to None when calling the
        # DatastoreTrace.

        if arg_extractor is None:
            index = None
        else:
            index = arg_extractor(*args, **kwargs)

        if prefix:
            operation = '%s.%s' % (prefix, name)
        else:
            operation = name

        transaction._nr_datastore_instance_info = (None, None, None)

        dt = DatastoreTrace(
                transaction=transaction,
                product='Elasticsearch',
                target=index,
                operation=operation
        )

        with dt:
            result = wrapped(*args, **kwargs)

            instance_info = transaction._nr_datastore_instance_info
            host, port_path_or_id, _ = instance_info

            dt.host = host
            dt.port_path_or_id = port_path_or_id

            return result

    if hasattr(owner, name):
        wrap_function_wrapper(owner, name, _nr_wrapper_Elasticsearch_method_)

_elasticsearch_client_methods = (
    ('abort_benchmark', None),
    ('benchmark', _extract_args_index),
    ('bulk', None),
    ('clear_scroll', None),
    ('count', _extract_args_index),
    ('count_percolate', _extract_args_index),
    ('create', _extract_args_index),
    ('delete', _extract_args_index),
    ('delete_by_query', _extract_args_index),
    ('delete_script', None),
    ('delete_template', None),
    ('exists', _extract_args_index),
    ('explain', _extract_args_index),
    ('get', _extract_args_index),
    ('get_script', None),
    ('get_source', _extract_args_index),
    ('get_template', None),
    ('index', _extract_args_index),
    ('info', None),
    ('list_benchmarks', _extract_args_index),
    ('mget', None),
    ('mlt', _extract_args_index),
    ('mpercolate', _extract_args_body_index),
    ('msearch', None),
    ('mtermvectors', None),
    ('percolate', _extract_args_index),
    ('ping', None),
    ('put_script', None),
    ('put_template', None),
    ('scroll', None),
    ('search', _extract_args_index),
    ('search_exists', _extract_args_index),
    ('search_shards', _extract_args_index),
    ('search_template', _extract_args_index),
    ('suggest', _extract_args_body_index),
    ('termvector', _extract_args_index),
    ('termvectors', None),
    ('update', _extract_args_index),
)

def instrument_elasticsearch_client(module):
    for name, arg_extractor in _elasticsearch_client_methods:
        wrap_elasticsearch_client_method(module.Elasticsearch, name,
                arg_extractor)

_elasticsearch_client_indices_methods = (
    ('analyze', _extract_args_index),
    ('clear_cache', _extract_args_index),
    ('close', _extract_args_index),
    ('create', _extract_args_index),
    ('delete', _extract_args_index),
    ('delete_alias', _extract_args_index),
    ('delete_mapping', _extract_args_index),
    ('delete_template', None),
    ('delete_warmer', _extract_args_index),
    ('exists', _extract_args_index),
    ('exists_alias', _extract_args_name_index),
    ('exists_template', None),
    ('exists_type', _extract_args_index),
    ('flush', _extract_args_index),
    ('get', _extract_args_index),
    ('get_alias', _extract_args_index),
    ('get_aliases', _extract_args_index),
    ('get_mapping', _extract_args_index),
    ('get_field_mapping', _extract_args_field_index),
    ('get_settings', _extract_args_index),
    ('get_template', None),
    ('get_upgrade', _extract_args_index),
    ('get_warmer', _extract_args_index),
    ('open', _extract_args_index),
    ('optimize', _extract_args_index),
    ('put_alias', _extract_args_name_index),
    ('put_mapping', _extract_args_doctype_body_index),
    ('put_settings', _extract_args_body_index),
    ('put_template', None),
    ('put_warmer', _extract_args_name_body_index),
    ('recovery', _extract_args_index),
    ('refresh', _extract_args_index),
    ('segments', _extract_args_index),
    ('snapshot_index', _extract_args_index),
    ('stats', _extract_args_index),
    ('status', _extract_args_index),
    ('update_aliases', None),
    ('upgrade', _extract_args_index),
    ('validate_query', _extract_args_index),
)

def instrument_elasticsearch_client_indices(module):
    for name, arg_extractor in _elasticsearch_client_indices_methods:
        wrap_elasticsearch_client_method(module.IndicesClient, name,
                arg_extractor, 'indices')

_elasticsearch_client_cat_methods = (
    ('aliases', None),
    ('allocation', None),
    ('count', _extract_args_index),
    ('fielddata', None),
    ('health', None),
    ('help', None),
    ('indices', _extract_args_index),
    ('master', None),
    ('nodes', None),
    ('pending_tasks', None),
    ('plugins', None),
    ('recovery', _extract_args_index),
    ('shards', _extract_args_index),
    ('segments', _extract_args_index),
    ('thread_pool', None),
)

def instrument_elasticsearch_client_cat(module):
    for name, arg_extractor in _elasticsearch_client_cat_methods:
        wrap_elasticsearch_client_method(module.CatClient, name,
                arg_extractor, 'cat')

_elasticsearch_client_cluster_methods = (
    ('get_settings', None),
    ('health', _extract_args_index),
    ('pending_tasks', None),
    ('put_settings', None),
    ('reroute', None),
    ('state', _extract_args_metric_index),
    ('stats', None),
)

def instrument_elasticsearch_client_cluster(module):
    for name, arg_extractor in _elasticsearch_client_cluster_methods:
        wrap_elasticsearch_client_method(module.ClusterClient, name,
                arg_extractor, 'cluster')

_elasticsearch_client_nodes_methods = (
    ('hot_threads', None),
    ('info', None),
    ('shutdown', None),
    ('stats', None),
)

def instrument_elasticsearch_client_nodes(module):
    for name, arg_extractor in _elasticsearch_client_nodes_methods:
        wrap_elasticsearch_client_method(module.NodesClient, name,
                arg_extractor, 'nodes')

_elasticsearch_client_snapshot_methods = (
    ('create', None),
    ('create_repository', None),
    ('delete', None),
    ('delete_repository', None),
    ('get', None),
    ('get_repository', None),
    ('restore', None),
    ('status', None),
    ('verify_repository', None),
)

def instrument_elasticsearch_client_snapshot(module):
    for name, arg_extractor in _elasticsearch_client_snapshot_methods:
        wrap_elasticsearch_client_method(module.SnapshotClient, name,
                arg_extractor, 'snapshot')

_elasticsearch_client_tasks_methods = (
    ('list', None),
    ('cancel', None),
    ('get', None),
)

def instrument_elasticsearch_client_tasks(module):
    for name, arg_extractor in _elasticsearch_client_tasks_methods:
        wrap_elasticsearch_client_method(module.TasksClient, name,
                arg_extractor, 'tasks')

_elasticsearch_client_ingest_methods = (
    ('get_pipeline', None),
    ('put_pipeline', None),
    ('delete_pipeline', None),
    ('simulate', None),
)

def instrument_elasticsearch_client_ingest(module):
    for name, arg_extractor in _elasticsearch_client_ingest_methods:
        wrap_elasticsearch_client_method(module.IngestClient, name,
                arg_extractor, 'ingest')

#
# Instrumentation to get Datastore Instance Information
#

def _nr_Connection__init__wrapper(wrapped, instance, args, kwargs):
    """Cache datastore instance info on Connection object"""

    def _bind_params(host='localhost', port=9200, *args, **kwargs):
        return host, port

    host, port = _bind_params(*args, **kwargs)
    port = str(port)
    instance._nr_host_port = (host, port)

    return wrapped(*args, **kwargs)

def instrument_elasticsearch_connection_base(module):
    wrap_function_wrapper(module.Connection, '__init__',
            _nr_Connection__init__wrapper)

def _nr_get_connection_wrapper(wrapped, instance, args, kwargs):
    """Read instance info from Connection and stash on Transaction."""

    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    conn = wrapped(*args, **kwargs)

    instance_info = (None, None, None)
    try:
        tracer_settings = transaction.settings.datastore_tracer

        if tracer_settings.instance_reporting.enabled:
            host, port_path_or_id = conn._nr_host_port
            instance_info = (host, port_path_or_id, None)
    except:
        instance_info = ('unknown', 'unknown', None)

    transaction._nr_datastore_instance_info = instance_info

    return conn

def instrument_elasticsearch_transport(module):
    wrap_function_wrapper(module.Transport, 'get_connection',
            _nr_get_connection_wrapper)



import newrelic.api.function_trace

_methods_1 = ['bgrewriteaof', 'bgsave', 'config_get', 'config_set',
              'dbsize', 'debug_object', 'delete', 'echo', 'flushall',
              'flushdb', 'info', 'lastsave', 'object', 'ping', 'save',
              'shutdown', 'slaveof', 'append', 'decr', 'exists',
              'expire', 'expireat', 'get', 'getbit', 'getset', 'incr',
              'keys', 'mget', 'mset', 'msetnx', 'move', 'persist',
              'randomkey', 'rename', 'renamenx', 'set', 'setbit',
              'setex', 'setnx', 'setrange', 'strlen', 'substr', 'ttl',
              'type', 'blpop', 'brpop', 'brpoplpush', 'lindex',
              'linsert', 'llen', 'lpop', 'lpush', 'lpushx', 'lrange',
              'lrem', 'lset', 'ltrim', 'rpop', 'rpoplpush', 'rpush',
              'rpushx', 'sort', 'sadd', 'scard', 'sdiff', 'sdiffstore',
              'sinter', 'sinterstore', 'sismember', 'smembers',
              'smove', 'spop', 'srandmember', 'srem', 'sunion',
              'sunionstore', 'zadd', 'zcard', 'zcount', 'zincrby',
              'zinterstore', 'zrange', 'zrangebyscore', 'zrank', 'zrem',
              'zremrangebyrank', 'zremrangebyscore', 'zrevrange',
              'zrevrangebyscore', 'zrevrank', 'zscore', 'zunionstore',
              'hdel', 'hexists', 'hget', 'hgetall', 'hincrby', 'hkeys',
              'hlen', 'hset', 'hsetnx', 'hmset', 'hmget', 'hvals',
              'publish']

_methods_2 = ['setex', 'lrem', 'zadd']

def instrument_redis_connection(module):

    newrelic.api.function_trace.wrap_function_trace(
        module, 'Connection.connect')

def instrument_redis_client(module):

    if hasattr(module, 'StrictRedis'):
        for method in _methods_1:
            if hasattr(module.StrictRedis, method):
                newrelic.api.function_trace.wrap_function_trace(
                        module, 'StrictRedis.%s' % method)
    else:
        for method in _methods_1:
            if hasattr(module.Redis, method):
                newrelic.api.function_trace.wrap_function_trace(
                        module, 'Redis.%s' % method)

    for method in _methods_2:
        if hasattr(module.Redis, method):
            newrelic.api.function_trace.wrap_function_trace(
                    module, 'Redis.%s' % method)



# -*- coding: utf-8 -*-

"""
pythoncompat
"""

from .packages import chardet

import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

#: Python 3.0.x
is_py30 = (is_py3 and _ver[1] == 0)

#: Python 3.1.x
is_py31 = (is_py3 and _ver[1] == 1)

#: Python 3.2.x
is_py32 = (is_py3 and _ver[1] == 2)

#: Python 3.3.x
is_py33 = (is_py3 and _ver[1] == 3)

#: Python 3.4.x
is_py34 = (is_py3 and _ver[1] == 4)

#: Python 2.7.x
is_py27 = (is_py2 and _ver[1] == 7)

#: Python 2.6.x
is_py26 = (is_py2 and _ver[1] == 6)

#: Python 2.5.x
is_py25 = (is_py2 and _ver[1] == 5)

#: Python 2.4.x
is_py24 = (is_py2 and _ver[1] == 4)   # I'm assuming this is not by choice.


# ---------
# Platforms
# ---------


# Syntax sugar.
_ver = sys.version.lower()

is_pypy = ('pypy' in _ver)
is_jython = ('jython' in _ver)
is_ironpython = ('iron' in _ver)

# Assume CPython, if nothing else.
is_cpython = not any((is_pypy, is_jython, is_ironpython))

# Windows-based system.
is_windows = 'win32' in str(sys.platform).lower()

# Standard Linux 2+ system.
is_linux = ('linux' in str(sys.platform).lower())
is_osx = ('darwin' in str(sys.platform).lower())
is_hpux = ('hpux' in str(sys.platform).lower())   # Complete guess.
is_solaris = ('solar==' in str(sys.platform).lower())   # Complete guess.

try:
    import simplejson as json
except (ImportError, SyntaxError):
    # simplejson does not support Python 3.2, it thows a SyntaxError
    # because of u'...' Unicode literals.
    import json

# ---------
# Specifics
# ---------

if is_py2:
    from urllib import quote, unquote, quote_plus, unquote_plus, urlencode, getproxies, proxy_bypass
    from urlparse import urlparse, urlunparse, urljoin, urlsplit, urldefrag
    from urllib2 import parse_http_list
    import cookielib
    from Cookie import Morsel
    from StringIO import StringIO
    from .packages.urllib3.packages.ordered_dict import OrderedDict

    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring
    numeric_types = (int, long, float)


elif is_py3:
    from urllib.parse import urlparse, urlunparse, urljoin, urlsplit, urlencode, quote, unquote, quote_plus, unquote_plus, urldefrag
    from urllib.request import parse_http_list, getproxies, proxy_bypass
    from http import cookiejar as cookielib
    from http.cookies import Morsel
    from io import StringIO
    from collections import OrderedDict

    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)



class AdaptiveSampler(object):
    def __init__(self, sampling_target):
        self.min_sampling_priority = 0.0

        # For the first harvest, collect a max of sampling_target number of
        # "sampled" transactions.
        self.sampling_target = sampling_target
        self.max_sampled = sampling_target
        self.max_ratio = 0.5

        self.transaction_count = 0
        self.sampled_count = 0

    def compute_sampled(self, priority):
        if self.sampled_count >= self.max_sampled:
            return False

        elif priority >= self.min_sampling_priority:
            self.sampled_count += 1

            # Determine if a backoff is required and compute it
            target = self.sampling_target
            if self.sampled_count > target:
                ratio = target / float(self.sampled_count)
                target = target ** ratio - target ** self.max_ratio
                target = max(0.0, target)

                self._update_min_priority(target)

            return True

        return False

    def reset(self, transaction_count):
        # For subsequent harvests, collect a max of twice the
        # self._sampling_target value.
        self.max_sampled = 2 * self.sampling_target
        self.transaction_count = transaction_count
        self.sampled_count = 0
        self._update_min_priority(self.sampling_target)

    def _update_min_priority(self, target):
        if self.transaction_count > target:
            sampling_ratio = float(target) / self.transaction_count
            self.min_sampling_priority = 1.0 - sampling_ratio
        else:
            self.min_sampling_priority = 0.0



import newrelic.api.memcache_trace

def instrument(module):

    if hasattr(module.Client, 'add'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.add', 'add')
    if hasattr(module.Client, 'append'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.append', 'replace')
    if hasattr(module.Client, 'decr'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.decr', 'decr')
    if hasattr(module.Client, 'delete'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.delete', 'delete')
    if hasattr(module.Client, 'delete_multi'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.delete_multi', 'delete')
    if hasattr(module.Client, 'get'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.get', 'get')
    if hasattr(module.Client, 'get_multi'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.get_multi', 'get')
    if hasattr(module.Client, 'incr'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.incr', 'incr')
    if hasattr(module.Client, 'incr_multi'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.incr_multi', 'incr')
    if hasattr(module.Client, 'prepend'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.prepend', 'replace')
    if hasattr(module.Client, 'replace'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.replace', 'replace')
    if hasattr(module.Client, 'set'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.set', 'set')
    if hasattr(module.Client, 'set_multi'):
        newrelic.api.memcache_trace.wrap_memcache_trace(
                module, 'Client.set_multi', 'set')



from newrelic.common.object_wrapper import wrap_function_wrapper


def _bind_make_coroutine_wrapper(func, *args, **kwargs):
    return func


def wrap_make_coroutine_wrapper(wrapped, instance, args, kwargs):
    coro = wrapped(*args, **kwargs)
    original_func = _bind_make_coroutine_wrapper(*args, **kwargs)
    coro.__wrapped__ = original_func
    coro.__tornado_coroutine__ = True
    return coro


def instrument_tornado_gen(module):
    import tornado
    if hasattr(tornado, 'version_info') and tornado.version_info < (4, 5):
        wrap_function_wrapper(module, '_make_coroutine_wrapper',
                wrap_make_coroutine_wrapper)



from collections import namedtuple

RootNode = namedtuple('RootNode',
        ['start_time', 'empty0', 'empty1', 'root', 'attributes'])

def root_start_time(root):
    return root.start_time * 1000.0

TraceNode = namedtuple('TraceNode',
        ['start_time', 'end_time', 'name', 'params', 'children', 'label'])

def node_start_time(root, node):
    return (node.start_time - root.start_time) * 1000.0

def node_end_time(root, node):
    return (node.end_time - root.start_time) * 1000.0



import sys

# These have been moved. They are retained here until all references to
# them are moved at which point will mark as deprecated to ensure users
# weren't using them directly.

from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import ObjectWrapper, wrap_object

# From Python 3.X. In older Python versions it fails if attributes do
# not exist and don't maintain a __wrapped__ attribute.

WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__doc__', '__annotations__')
WRAPPER_UPDATES = ('__dict__',)

def update_wrapper(wrapper,
                   wrapped,
                   assigned = WRAPPER_ASSIGNMENTS,
                   updated = WRAPPER_UPDATES):
    """Update a wrapper function to look like the wrapped function

       wrapper is the function to be updated
       wrapped is the original function
       assigned is a tuple naming the attributes assigned directly
       from the wrapped function to the wrapper function (defaults to
       functools.WRAPPER_ASSIGNMENTS)
       updated is a tuple naming the attributes of the wrapper that
       are updated with the corresponding attribute from the wrapped
       function (defaults to functools.WRAPPER_UPDATES)
    """
    wrapper.__wrapped__ = wrapped
    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)
    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper



from newrelic.hooks.framework_tornado_r3.util import (
        retrieve_current_transaction, create_transaction_aware_fxn)
from newrelic.common.object_wrapper import wrap_function_wrapper


def _nr_wrapper_Future_add_done_callback(wrapped, instance, args, kwargs):
    def _fxn_arg_extractor(fn, *args, **kwargs):
        return fn

    fxn = _fxn_arg_extractor(*args, **kwargs)

    should_trace = not hasattr(fxn, '_nr_last_object')

    # In Tornado, the coroutine runs inside of a lambda or a closure. In this
    # case, we don't know the name of the actual function until it's run. The
    # name is passed as an attribute on the future.
    if hasattr(instance, '_nr_coroutine_name'):
        old_fxn = fxn

        def new_fxn(*_args, **_kwargs):
            return old_fxn(*_args, **_kwargs)

        fxn = new_fxn
        fxn._nr_coroutine_name = instance._nr_coroutine_name

        # Clear name off of future in case the future is reused
        delattr(instance, '_nr_coroutine_name')

    transaction_aware_fxn = create_transaction_aware_fxn(fxn,
            should_trace=should_trace)

    # If transaction_aware_fxn is None then it is already wrapped, or the fxn
    # is None.
    if transaction_aware_fxn is None:
        return wrapped(*args, **kwargs)

    transaction = retrieve_current_transaction()

    transaction_aware_fxn._nr_transaction = transaction

    # We replace the function we call in the callback with the transaction
    # aware version of the function.
    if len(args) > 0:
        args = list(args)
        args[0] = transaction_aware_fxn
    else:
        # Keyword argument name for the callable function is 'fn'.
        kwargs['fn'] = transaction_aware_fxn

    return wrapped(*args, **kwargs)


def instrument_concurrent(module):

    # This is for instrumenting both tornado futures and python native futures

    wrap_function_wrapper(module, 'Future.add_done_callback',
            _nr_wrapper_Future_add_done_callback)
